
/*************************************************************************************************
Simple List Class

Copyright (C) 2009 Ezequiel Gastn Miravalles

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*************************************************************************************************/

/*************************************************************************************************
Software: Simple List Class
Author: Ezequiel Miravalles
Website: http://www.neoegm.com/tech/programming/c-cpp/simple-list-class/
License: GNU GPL v3 (attached, read above)
Last modification: 2009-08-18
*************************************************************************************************/

#pragma once

#include <iostream>

#include "cppMemDbg.h"

template <class TYPE>
class List
{
	struct Nodo
	{
		Nodo() { m_pNext = NULL; }

		TYPE m_tData;
		Nodo *m_pNext;
	};

	Nodo *m_pnHeader;

	//For optimization
	int m_nCount;
	Nodo *m_pnLast;
	//-----------------

public:
	//Construction and destruction
	List() { Init(); }
	~List() { FreeList(); }
	
	List(const List& rlList) { Init(); operator=(rlList); }
	
	//Assignment operator
	List& operator=(const List& rlList);

	//Information
	int Length() { return m_nCount; }
	bool Empty() { return m_nCount==0; }

	//Element managing
	int Add(TYPE& rtData);
	TYPE* Elem(int nPos);
	bool Delete(int nPos);
	void DeleteAll();
	
	//Search
	int Find(TYPE& rItem, int nStartAt = 0);

	//Operadores
	TYPE& operator[](int nPos) { return *Elem(nPos); }		//Elem
	int operator<<(TYPE& rdData) { return Add(rdData); }	//Add

protected:
	void FreeList();
	void Init() { m_pnHeader = NULL; m_pnLast = NULL; m_nCount = 0; }
};

template <class TYPE>
List<TYPE>& List<TYPE>::operator=(const List<TYPE>& rlList)
{
	DeleteAll();
	
	//Copy the entire list adding each element one by one
	for (Nodo *pnTemp = rlList.m_pnHeader; pnTemp; pnTemp = pnTemp->m_pNext)
		Add(pnTemp->m_tData);
	
	return *this;
}

template <class TYPE>
TYPE* List<TYPE>::Elem(int nPos)
{
	if (nPos >= m_nCount)
		return NULL;

	Nodo *pnNodo = m_pnHeader;

	for (int i = 0; i < nPos; i++)
		pnNodo = pnNodo?pnNodo->m_pNext:NULL;	//Just in case

	return &pnNodo->m_tData;
}

template <class TYPE>
void List<TYPE>::FreeList()
{
	while (m_pnHeader)
	{
		Nodo *pnNext = m_pnHeader->m_pNext;
		if (m_pnHeader)
			delete m_pnHeader;
		m_pnHeader = pnNext;
	}
}

template <class TYPE>
int List<TYPE>::Add(TYPE& rtData)
{
	Nodo *pnNodo = new Nodo;

	if (!pnNodo)	//Could not allocate
		return -1;

	pnNodo->m_tData = rtData;

	if (m_pnLast)
		m_pnLast->m_pNext = pnNodo;
	else
		m_pnHeader = pnNodo;

	m_pnLast = pnNodo;

	return m_nCount++;
}

template <class TYPE>
bool List<TYPE>::Delete(int nPos)
{
	if (nPos >= m_nCount)
		return NULL;

	if (nPos == 0)		//Delete the first one
	{
		Nodo *pnTemp = m_pnHeader->m_pNext;

		if (m_pnHeader)
			delete m_pnHeader;
		m_pnHeader = pnTemp;

		m_nCount--;

		if (!m_nCount)
			m_pnLast = NULL;

		return true;
	}

	Nodo *pnNodo = m_pnHeader;

	for (int i = 0; i < nPos+1; i++)
	{
		if (i+1 == nPos)
		{
			if (!pnNodo->m_pNext)				//Security verification
				return false;

			Nodo *pnTemp = pnNodo->m_pNext->m_pNext;

			if (pnNodo->m_pNext)
				delete pnNodo->m_pNext;
			pnNodo->m_pNext = pnTemp;

			m_nCount--;

			if (i+2 == m_nCount)	//Last node
				m_pnLast = pnNodo;

			return true;
		}

		pnNodo = pnNodo?pnNodo->m_pNext:NULL;	//Security verification

		if (!pnNodo)							//Security verification
			return false;
	}

	return false;
}

template <class TYPE>
void List<TYPE>::DeleteAll()
{
	FreeList();

	m_nCount = 0;
	m_pnLast = NULL;
}

template <class TYPE>
int List<TYPE>::Find(TYPE& rItem, int nStartAt /*= 0*/)
{
	for (int i = 0; i < Length(); i++)
	{
		if (i < nStartAt)
			continue;
		
		TYPE *tData = Elem(i);

		if (tData && *tData == rItem)	//The == operator must be defined for TYPE
			return i;
	}

	return -1;
}

template <class TYPE>
std::ostream& operator<<(std::ostream& oStream, List<TYPE>& rlList)
{
	for (int i = 0; i < rlList.Length(); i++)
	{
		TYPE *tData = rlList.Elem(i);

		if (tData)	//Just in case
		{
			oStream << *tData;		//The << operator must be defined for TYPE
			oStream << std::endl;
		}
	}

	return oStream;
}
